/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (initial author of incremfa)                          *
 *   - INRIA (modifications/diffs in SVN)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef DECOMPOSITION_LEVEL_H
#define DECOMPOSITION_LEVEL_H

#include "CyclicArray.h"
#include "Wavelet.h"

/** Fast DWT update algo is based on recursive half-window overlapping levels
    One of these is the active one, corresponding to the even data frames
    The other one has the last data point waiting around. When a new point arrives,
    it completes this odd frame into an even one, and the buffers are swapped.
    The new point was also added to the previously active buffer, which becomes odd-framed
    with the last point in the waiting state.

    Note this doesn't introduce an additional delay in the DWT coeficients: the last
    point is always taken in account in the new "active" buffer. Of course, the normal delay
    applies for getting low-frequency components out of an instantaneous measure, but
    this is completely unrelated to this algorithm. Recursively, a whole new DWT tree of active
    buffers is updated in O(L) constant time (it does not depend on the data buffer size).
    Another way to look at this algorithm is to consider there are 2^L windows maintained,
    in parallel, ending at the last 2^L consecutive indices. However, most of these frames
    share common points: for example two consecutive even frames of the first decomposition
    of the data have N/2 points each, among which N/2 - 1 are common points.

    The buffer pointers are kept around for easy pointer swapping on data update. Therefore,
    nextLevel and nextLevelHP always point to the current, valid, decomposition.
*/
class DecompositionLevel {
public:
    // See class description
    DecompositionLevel* nextLevelWaiting;

protected:

    int newDataUpdateIdx;

    // pointer to member type
    typedef DecompositionLevel* (DecompositionLevel::* NewDataFunctionPtr)(double*); 
    NewDataFunctionPtr newDataUpdatePtr;

    /** Updates this buffer with new data. As this is a cyclic buffer, the oldest one is discarded.
        The waiting and active next level buffers are updated & swapped as necessary.
        This operation is constant-time O(L) with L number of decomposition levels.
        Warning: this version is designed for better performance and assumes the array is
        full (use the other method for the initialization phase).
     */
    DecompositionLevel* newDataUpdateFullRegime(double* lphp);

    /** Updates this buffer with new data. As this is a cyclic buffer, the oldest one is discarded.
        The waiting and active next level buffers are updated & swapped as necessary.
        This operation is constant-time O(L) with L number of decomposition levels
        This method could be used in both Initialization phase and for the full regime,
        but the other method will be more efficient for the full regime. Hence
        this function automatically switches to the other one when ready
     */
    DecompositionLevel* newDataUpdateInitPhase(double* lphp);

    /// common wavelet instance shared amongst all decomposition levels
    Wavelet* wavelet;

public:

    /** Create a tree of depth nlevel, with ncoefmin coefficients in
        the highest level low-pass filer (that is, the level with
        larger scale and thus lesser number of coefficients)
        Each level uses up ws extra coefficients where ws is the wavelet
        support size: ws=4 for Daubechies W4 for example.

        This is due to the fact we don't wan't to make the data cyclic
        or extend the last values. This would create harmful artifacts, and
        complicate things in our case of cyclic buffers.

        However, for our usage, the data is actually a history array, and we
        don't care if the oldest values are not reconstructed. On the other
        hand, that half the discarded values are new ones may be problemetic
        for the delay introduced between the arrival of a new data and it's full
        inclusion in the spectrum computations.

        Computes capacity before recursion (alternative: use a class factory)
    */
    DecompositionLevel(int nlevel, int ncoefmin, Wavelet* w);
    ~DecompositionLevel();

    /// Reset this level. Use this instead of destroying/reallocating this object
    /// with the same parameters: the result is the same, but this function is much
    /// faster
    void reset();

    /// next level data (effective, not waiting)
    DecompositionLevel* nextLevel;
    /// previous level always effective
    DecompositionLevel* prevLevel;

    /// The coefficients
    CyclicArray lowPass;
    CyclicArray highPass;

    /// Maintain wavelet leaders on the fly (derived from highPass and levels below)
    CyclicArray leaders, leadersD1;

    /** Updates this buffer with new data. As this is a cyclic buffer, the oldest one is discarded.
        The waiting and active next level buffers are updated & swapped as necessary.
        This operation is constant-time O(L) with L number of decomposition levels.
     */
    inline DecompositionLevel* newDataUpdate(double* lphp) {
        return (this->*newDataUpdatePtr)(lphp);
    }

    /// Power sum of the delta at this scale for different q (double array[n_q])
    double *powerSum;
    /// How many terms were added so far in the Power Sum
    unsigned long pscount;
    /// last term need estimation of dyadic interval k+1 => only approximate so far
    double *powerSumLastTerm;

    /// create the power sum arrays with the given size
    void createPowerSum(int nq);

// Define this symbol if you wish perfect match between the memory and CPU versions.
// Unfortunately the expense of using long double nearly destroy the speed advantage
// of MEM vs CPU. According to your system, it might then be more advantageous to 
// stick to the CPU version!
#ifndef MEM_CPU_VERSIONS_PERFECT_MATCH
    /// According to user choice the first (2^L - 2^s) values can be recomputed (uses CPU,
    /// no mem) or re-used from last update (uses less CPU, but this quickly uses a lot of
    /// memory since there are 2^s decomposition levels at scale s!)
    /// These arrays are only allocated when the user decided to use Memory, they are not
    /// allocated when using CPU. See Analyzer.
    double* cachedNewDataDelta;
    /// The old data cache is used only when discarding old data
    double* cachedOldDataDelta;
#else
    long double* cachedNewDataDelta;
    long double* cachedOldDataDelta;
#endif

};

#endif
