/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (initial author of incremfa)                          *
 *   - INRIA (modifications/diffs in SVN)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <iomanip>
#include <map>
#include <list>
using namespace std;

#include <math.h>

#include "MultifractalAnalyzer.h"
#include "DensityEstimator.h"

#ifndef QSETTINGS
#include "QSettings.h"
#else
#include QSETTINGS
#endif

// helper called by the constructors
void MultifractalAnalyzer::constructor(int nlevel, int ncoefmin, Wavelet* w, double* wght, int navg, bool discardOlderData, bool monitorCorrelation) {

    wavelet = w->ref();
    this->discardOlderData = discardOlderData;
    this->monitorCorrelation = monitorCorrelation;

    localHolderEstimatesUseLeaders3K = false;

    // make local copy of weights
    weights = new double[nlevel];
    memcpy(weights, wght, nlevel*sizeof(double));

    nframes = 1 << nlevel;

    delay = (nframes - 1) * (w->support - 2);

    nextLevel = new DecompositionLevel(nlevel-1, ncoefmin, w);
    nextLevelWaiting = new DecompositionLevel(nlevel-1, ncoefmin, w);
    nextLevel->createPowerSum(NUM_Q);
    nextLevelWaiting->createPowerSum(NUM_Q);

    this->nlevel = nlevel;

    newDataDWTIdx = data.capacity() - wavelet->support;

    // Start with the initialization version
    // Then switch to full-buffer regime when this happens
    newDataDWTPtr = &MultifractalAnalyzer::newDataDWTInitPhase;

    // hq re-initialized each use, except when averaging over all frames
    hq = new double[NUM_Q];
    memset(hq, 0, NUM_Q * sizeof(double));
    if (monitorCorrelation) {
        squaredCorrelation = new double[NUM_Q];
        squaredCorrelationSum = new double[NUM_Q];
        memset(squaredCorrelation, 0, NUM_Q * sizeof(double));
        memset(squaredCorrelationSum, 0, NUM_Q * sizeof(double));
    } else squaredCorrelation =0;

    // Compute constants for H(q) computations from power sums

    // for the init phase, use effective number of non-empty levels
    powerSumToHqConstant = new double*[nlevel+1];
    powerSumToHqConstant[0] = 0;
    correlationCoefficientDenominatorTerm1 = new double[nlevel+1];
    correlationCoefficientDenominatorTerm1[0] = 0;
    meanX = new double[nlevel+1];
    meanX[0] = 0;
    sumOfWeightsInv = new double[nlevel+1];
    sumOfWeightsInv[0] = 0;
    for (effectiveNLevel=1; effectiveNLevel<=nlevel; ++effectiveNLevel) {
        powerSumToHqConstant[effectiveNLevel] = new double[effectiveNLevel+1];
        powerSumToHqConstant[effectiveNLevel][0] = 0;

        // Now use weighted least squares fitting formula
        // TODO: could commonalize these intermediary computations...
        double m1x = 0.0;
        double m2x = 0.0;
        double sw = 0.0;
        for (int l=0; l<effectiveNLevel; ++l) {
            double x = l+1;
            m1x += x * weights[l];
            m2x += x*x * weights[l];
            sw += weights[l];
        }
        double commonConst = sw * m2x - m1x * m1x;

        // variables used for correlation coefficient computation
        correlationCoefficientDenominatorTerm1[effectiveNLevel] = m2x - m1x * m1x / sw;
        meanX[effectiveNLevel] = m1x / sw;
        sumOfWeightsInv[effectiveNLevel] = 1.0 / sw;

        for (int l=0; l<effectiveNLevel; ++l) {
            powerSumToHqConstant[effectiveNLevel][l+1] = weights[l] * (sw * (l+1) - m1x) / commonConst;
        }
    }

    // With uniform weighting the result of the computation below should be -0.5 * GET_Q_FOR_INDEX(i) for each q
    // With non-uniform weighting...
    HQEnergyCorrection = new double*[effectiveNLevel+1];
    HQEnergyCorrection[0] = 0;
    for (effectiveNLevel=1; effectiveNLevel<=nlevel; ++effectiveNLevel) {
        HQEnergyCorrection[effectiveNLevel] = new double[NUM_Q];
        for (int i=0; i<NUM_Q; ++i) {
            HQEnergyCorrection[effectiveNLevel][i] = 0;
            for (int scale = 1; scale<=effectiveNLevel; ++scale) {
                HQEnergyCorrection[effectiveNLevel][i] += powerSumToHqConstant[effectiveNLevel][scale] * (-0.5) *scale*GET_Q_FOR_INDEX(i);
            }
        }
    }

    // no guarantee the value q=1 is even included in the QSettings, so...
    // nlevel is always effective in the LH case
    LHEnergyCorrection = 0;
    for (int scale = 1; scale<=nlevel; ++scale) {
        LHEnergyCorrection += powerSumToHqConstant[nlevel][scale] * (-0.5) * scale;
    }

    effectiveNLevel = 0; // to begin with
}

MultifractalAnalyzer::~MultifractalAnalyzer() {
    // cleanup hq
    delete [] hq;

    // cleanup constants
    for (int s=1; s<=nlevel; ++s) {
        delete [] powerSumToHqConstant[s];
    }
    delete [] powerSumToHqConstant;

    delete [] correlationCoefficientDenominatorTerm1;
    delete [] meanX;
    delete [] sumOfWeightsInv;


    delete [] weights;

    // cleanup levels (recursive)
    delete nextLevel;
    delete nextLevelWaiting;

    if (monitorCorrelation) {
        delete [] squaredCorrelation;
        delete [] squaredCorrelationSum;
    }

    // cleanup history (if needed)
    for (int i=0; i<hqHistory.size(); ++i) delete [] hqHistory[i];

    // cleanup Correlation history (if needed)
    for (int i=0; i<correlationHistory.size(); ++i) delete [] correlationHistory[i];

    // release shared instance
    wavelet->unref();
}

void MultifractalAnalyzer::reset() {

    // clear histories
    for (int i=0; i<hqHistory.size(); ++i) delete [] hqHistory[i];
    for (int i=0; i<correlationHistory.size(); ++i) delete [] correlationHistory[i];
    hqHistory.size() = 0;
    correlationHistory.size() = 0;

    // hq re-initialized too
    memset(hq, 0, NUM_Q * sizeof(double));
    if (squaredCorrelation) memset(squaredCorrelation, 0, NUM_Q * sizeof(double));

    // Clear data
    data.size() = 0;

    // Start with the initialization version
    // Then switch to full-buffer regime when this happens
    newDataDWTPtr = &MultifractalAnalyzer::newDataDWTInitPhase;

    // recursive reset levels
    if (nextLevel) nextLevel->reset();
    if (nextLevelWaiting) nextLevelWaiting->reset();
}

MultifractalAnalyzer::MultifractalAnalyzer(int nlevel, int ncoefmin, Wavelet* w, double* weights, int navg, bool discardOlderData, bool monitorCorrelation) : hqHistory(navg), correlationHistory(navg), localHolderExponentsSum((1 << nlevel) * ncoefmin), localHolderExponentsNum((1 << nlevel) * ncoefmin), data((1 << nlevel) * ncoefmin + (w->support - 2) * ((1 << nlevel) - 1)) {

    // Use user-specified weights for exponential fit
    constructor(nlevel, ncoefmin, w, weights, navg, discardOlderData, monitorCorrelation);
}

MultifractalAnalyzer::MultifractalAnalyzer(int nlevel, int ncoefmin, Wavelet* w, int navg, bool discardOlderData, bool monitorCorrelation) : hqHistory(navg), correlationHistory(navg), localHolderExponentsSum((1 << nlevel) * ncoefmin), localHolderExponentsNum((1 << nlevel) * ncoefmin), data((1 << nlevel) * ncoefmin + (w->support - 2) * ((1 << nlevel) - 1)) {

    // Use weighting in least squares estimate for exponential fit y = A * exp(B*x)   (eq.1)
    // or equivalently log y = B*x + log A     (eq.2)
    // Here we want to estimate B = h(q), with x = s * log(2). We don't care for A, so using the powerSums instead
    // of the full F(s) = (1/N * power_sum) ^ q is possible. Eq. 2 shows we also need to divide by q to get
    // a correct B back  (and similarly adding log N would be needed for A, but we don't care for A).
    // Simple least squares (no weighting) minimizes sum_over_all_(x,y)[ (log y - (B*x + log A))^2 ]
    // Unfortunately this gives more influence to lower scales due to the log transform
    // A possible way is to give all (x,y) a roughly identical influence by using the y as weights (as they
    // are close to the final exponential) : min sum_over_all_(x,y)[ y * (log y - (B*x + log A))^2 ]
    // Unfortunately this is still not quite correct, since y is only an estimate for the fitted exponential
    // and there may be large variations giving too much weight to a y (hence to that variation...).
    // Nevertheless, this is still much better than not using any weight at all. Other extentions would perhaps
    // consider using the estimated value of B to get a better weighting scheme than y, and iterate till
    // convergence.
    // In our case, using the y as weights would prevent any pre-computation. This is unfortunate as we don't
    // wan't to sacrifice too much speed for a slight increase in precision.
    // However, an exponential weigthing scheme does not cost more than unweighted scheme thanks to precomputation
    // => we might as well use it
    double weights[nlevel];
    for (int s=1; s<=nlevel; ++s) {
        weights[s-1] = 1;//sqrt((double)(1 << s));
    }
    weights[0]=0;
    weights[nlevel-1]=0;

    constructor(nlevel, ncoefmin, w, weights, navg, discardOlderData, monitorCorrelation);

}

/// Discrete Wavelet Transform the data into the next level
/// low and high band coefficients
void MultifractalAnalyzer::transform(CyclicArray& data, DecompositionLevel* level) {
    // Boundary coefficients won't be reconstructed,
    // there is no way around that if we don't wan't to introduce
    // boundary effects (which would have too much influence on
    // the scale/frequency analysis)
    const int max = data.size() - wavelet->support;
    double lphp[2];
    for (int i=0; i<=max; i+=2) {
        wavelet->convolve(data, i, lphp);
        level->lowPass[i>>1] = lphp[0];
        level->highPass[i>>1] = lphp[1];
    }
}

/// restore data from next level by inverse DWT
void MultifractalAnalyzer::restore(DecompositionLevel* level, CyclicArray& data) {
    const int max = level->lowPass.size()-(wavelet->support-2)/2;
    // Boundary coefficients are not reconstructed
    int idx = 0;
    for (; idx < wavelet->support-2; ++idx) data[idx] = 0;
    double evenodd[2];
    for (int i=0; i<max; ++i) {
        wavelet->reconstruct(level->lowPass, level->highPass, i, evenodd);
        data[idx++] = evenodd[0];
        data[idx++] = evenodd[1];
    }
    // other half of the discarded coefficients
    while (idx < data.size()) data[idx++] = 0;
}

DecompositionLevel* MultifractalAnalyzer::newDataDWTInitPhase(double newData) {
    if (data.isFull()) {
        // remove old data
        data.pop_front();
        // Switch to full buffer version for next time
        newDataDWTPtr = &MultifractalAnalyzer::newDataDWTFullRegime;
    }
    // add new data
    data.push_back(newData);
    // Swap buffers
    DecompositionLevel* tmplp = nextLevelWaiting;
    nextLevelWaiting = nextLevel;
    nextLevel = tmplp;
    // update DWT coefficients for feeding next level recursively
    const int idx = data.size() - wavelet->support;
    if (idx<0) return 0;
    double lphp[2];
    wavelet->convolve(data, idx, lphp);
    return nextLevel->newDataUpdate(lphp);
}

DecompositionLevel* MultifractalAnalyzer::newDataDWTFullRegime(double newData) {
    // add new data
    data.push_cyclic(newData);
    // Swap buffers
    DecompositionLevel* tmplp = nextLevelWaiting;
    nextLevelWaiting = nextLevel;
    nextLevel = tmplp;
    // update DWT coefficients for feeding next level recursively
    double lphp[2];
    wavelet->convolve(data, newDataDWTIdx, lphp);
    return nextLevel->newDataUpdate(lphp);
}



int MultifractalAnalyzer::newData(double data, bool updatePowerSums, bool updateLocalHolder) {

    // PART 1: Remove old point contribution to the MFA spectrum (old frame)
    //         This is done except if the user asked to compute over the whole
    //         series, from the beginning. This whole series mode may be
    //         appropriate for stationary series as it averages over more data.
    //         However keeping all data is inappropriate for non-stationary series.
    //--------

    if (discardOlderData) {

        if (updatePowerSums) for (DecompositionLevel *level = nextLevel; level; level = level->nextLevel) {
            // Stop if a level is not ready. By construction, upper levels aren't either
            if (!level->leaders.isFull()) break;

            double term = fabs(level->leaders[0]);

            // Process positive values
            minusExponentiationPos(term, level->powerSum);
#ifdef HAS_NEGATIVE_Q
            // Process negative values
            minusExponentiationNeg(1.0 / term, level->powerSum);
#endif
            // really bad conditionning: tell user and provoke full recomputation
            for (int i=0; i<NUM_Q; ++i) {
                if (level->powerSum[i]<0) {
                    //cout << "Too many numerical errors. Please process full recomputations from time to time." << endl;
                    level->powerSum[i]=0;//throw "error";
                }
            }
            //cout << endl;
            --level->pscount;
        }

        if (updateLocalHolder && localHolderExponentsSum.isFull()) {
            localHolderExponentsSum.pop_front();
            localHolderExponentsNum.pop_front();
        }
    }

    // PART 2: Update the DWT and the current frame power sum with the new data
    //--------
    DecompositionLevel* lastLevel = newDataDWT(data);

    if (updatePowerSums) {
        effectiveNLevel = 0;
        for (DecompositionLevel *level = nextLevel; level; level = level->nextLevel) {
            // running sum does NOT include last term as wavelet leader estimation is only complete at next update
            // must wait for dyadic intervals k-1, k, AND k+1 which is not available yet
            if (level->leaders.size()<2) break;
            double term = fabs(level->leaders[level->leaders.size()-2]);
            plusExponentiationPos(term, level->powerSum);
#ifdef HAS_NEGATIVE_Q
            plusExponentiationNeg(1.0 / term, level->powerSum);
#endif
            ++level->pscount;
        }

        for (DecompositionLevel *level = nextLevel; level; level = level->nextLevel) {
            // BUT current estimate might be used ASAP in estimation
            if (level->leaders.size()<1) break;
            double term = fabs(level->leaders[level->leaders.size()-1]);
            memset(level->powerSumLastTerm, 0, NUM_Q * sizeof(double));
            plusExponentiationPos(term, level->powerSumLastTerm);
#ifdef HAS_NEGATIVE_Q
            plusExponentiationNeg(1.0 / term, level->powerSumLastTerm);
#endif
            ++effectiveNLevel;
        }
    }

    if (updateLocalHolder) {
        if (!lastLevel) return 0;
        if (lastLevel->nextLevel) return 0; // need to be max level

        // leaders and leadersD1 have the same size
        if (lastLevel->leaders.size() < (localHolderEstimatesUseLeaders3K?2:1) ) return 1;

        // Initialize a new value so we can add estimates across frames
        localHolderExponentsSum.push_back(0.0);
        localHolderExponentsNum.push_back(0);

        int baseidx = (lastLevel->leaders.size()-(localHolderEstimatesUseLeaders3K?2:1))*nframes;

cout << "baseidx = " << baseidx<<endl;
cout << "lastLevel->leaders.size() = " << lastLevel->leaders.size() << endl;
cout << "localHolderExponentsNum.size() = " << localHolderExponentsNum.size() << endl;
cout << "nextLevel->data.lowPass() = " << nextLevel->lowPass.size() << endl;


        // TODO manage exponents per frame, do not mix between frames? if (hqHistory.capacity()!=1)
        for (int frame = 0; frame < nframes; ++frame) {
            double localH = LHEnergyCorrection;
            int fidx = baseidx + frame;

            DecompositionLevel *level = nextLevel;
            for (int scale = 1; scale<=nlevel; ++scale) {
                fidx /= 2;
                double wl = fabs((localHolderEstimatesUseLeaders3K?level->leaders:level->leadersD1)[fidx]);

//if (wl==0) {cout << "Error: null leader!" << endl;}
if (wl==0) break;
                // OK, reuse the slope conversion constant, but we want negative slope here
                // so subtracts the values
    //#ifdef NO_YLOG2X_ASM
                localH += powerSumToHqConstant[nlevel][scale] * log2(wl);
    /*#else
                asm(
                "fyl2x\n\t"    // powerSumToHqConstant[effectiveNLevel][scale] * log2(wl) in one instruction
                "fsubrl %[lh]" // loads localH and remove previous value from it
                : "=t" (localH)
                : "u" (powerSumToHqConstant[nlevel][scale]), "0" (wl), [lh] "m" (localH)
                : "st(1)"
                );
    #endif
    */
                level = level->nextLevel;
            }
//if (!isnormal(localH)) {cout << "Error: NaN local holder exponent!" << endl;}
            // Use 0.0 so it can be added later on, NaN cannot
            if (!isnormal(localH)) localH = 0.0;

            // aggregate over frames, this is a property of the data, not the frame
            localHolderExponentsSum[baseidx + frame] += localH;
            localHolderExponentsNum[baseidx + frame] += 1;
        }

    }

    // take care of initialization phase.
    if (!lastLevel) return 0;
    if (!lastLevel->leaders.isFull()) return 1;
    return 2;
}


// Code inspired from the matlab file fast_univariate_bandwidth_estimate_STEPI.m by Vikas C. Raykar, LGPL v2.1+
// Yes, even some comments are kept. Let's give credit where it's due.
double MultifractalAnalyzer::computeOptimalKernelWidth(double epsilon) {

    vector<double> exponents(localHolderExponentsNum.size());
    for (unsigned int i=0; i<exponents.size(); ++i) exponents[i] = localHolderExponentsSum[i] / localHolderExponentsNum[i];

    return ::computeOptimalKernelWidth(exponents, epsilon);
}

double MultifractalAnalyzer::estimateLocalHolderDensity(std::vector<double>& alpha, std::vector<double>& Dalpha, double kernelWidth, double epsilon) {
    Dalpha.resize(alpha.size());
    if (Dalpha.empty()) return 0.0;

    vector<double> exponents(localHolderExponentsNum.size());
    for (unsigned int i=0; i<exponents.size(); ++i) exponents[i] = localHolderExponentsSum[i] / localHolderExponentsNum[i];

    return estimateDensity(exponents, alpha, Dalpha, kernelWidth, epsilon);
}

double MultifractalAnalyzer::getLocalHolderSpectrum(std::vector<double>& alpha, std::vector<double>& Dalpha, int numAlpha, double kernelWidth, double epsilon) {
    if (numAlpha<=0) numAlpha = localHolderExponentsSum.size();
    alpha.resize(numAlpha);
    Dalpha.resize(numAlpha);
    if (numAlpha==0) return 0.0;

    vector<double> exponents(localHolderExponentsNum.size());
    for (unsigned int i=0; i<exponents.size(); ++i) exponents[i] = localHolderExponentsSum[i] / localHolderExponentsNum[i];

    double smin = exponents[0];
    double smax = smin;
    for (unsigned int i=1; i<exponents.size(); ++i) {
        if (exponents[i] > smax) smax = exponents[i];
        if (exponents[i] < smin) smin = exponents[i];
    }

    if (numAlpha==1) alpha[0] = (smin + smax) * 0.5;
    else for (int i=0; i<numAlpha; ++i) {
        alpha[i] = smin + i * (smax - smin) / (numAlpha - 1.0);
    }

    return estimateDensity(exponents, alpha, Dalpha, kernelWidth, epsilon);
}

void MultifractalAnalyzer::getLocalHolderExponentsHistogram(std::vector<double>& alpha, std::vector<int>& histogram, int numBins) {
    if (numBins<0) numBins = 0;
    alpha.resize(numBins);
    histogram.resize(numBins);
    if (numBins==0) return;

    vector<double> exponents(localHolderExponentsNum.size());
    for (unsigned int i=0; i<exponents.size(); ++i) exponents[i] = localHolderExponentsSum[i] / localHolderExponentsNum[i];

    double smin = exponents[0];
    double smax = smin;
    for (unsigned int i=1; i<exponents.size(); ++i) {
        if (exponents[i] > smax) smax = exponents[i];
        if (exponents[i] < smin) smin = exponents[i];
    }

    if (numBins==1) {
        alpha[0] = (smin + smax) * 0.5;
        histogram[0] = 1;
        return;
    }

    // extreme alpha centered on extreme bins, then even spacing
    double binSize = (smax - smin) / (numBins - 1.0);
    double bstart = smin - binSize*0.5;
    double invBinSize = 1.0 / binSize;

    for (int i=0; i<numBins; ++i) {
        alpha[i] = smin + i*binSize;
        histogram[i] = 0;
    }

    int N = exponents.size();
    for (int i=0; i<N; ++i) {
        int bidx = (int)floor((exponents[i] - bstart) * invBinSize);
        if (bidx<0) bidx = 0;
        if (bidx>=numBins) bidx = numBins;
        ++histogram[bidx];
    }

}

void MultifractalAnalyzer::getLegendreTransformSpectrum(std::map<double,double>& alpha_D, bool secondOrderLegendreApproximation) {

    // PART 3: Recompute the MFA spectrum
    //--------

    double * workHQ;
    double * workSquaredCorrelation;

    if (hqHistory.capacity()==1) workHQ = hq;
    else {
        if (hqHistory.isFull()) {
            workHQ = hqHistory[0];
            for (int i=0; i<NUM_Q; ++i) hq[i] -= workHQ[i];
            hqHistory.push_cyclic(workHQ); //pushes itself at the same place!
        } else {
            hqHistory.push_back(workHQ = new double[NUM_Q]);
        }
    }

    // no correlation => simpler form
    if (!monitorCorrelation) {

        // We deal here with non-power-of-two windows: no artificial wrapping, etc = not all data may be computed
        // As a result the power sums are first computed per element, so they are comparable between scales.
        //    energy correction: the wavelet coefficients are normalised with sqrt(2) => pow(sqrt(2), scale) energy
        //    total energy for scale s = num_elem * pow(2.0, 0.5*s)
        // The power sum is divided here by the number of elements. The log of the other energy term is pre-computed in HQEnergyCorrection.
        // Alternative form (without HQEnergyCorrection)
        //            workHQ[i] += powerSumToHqConstant[scale][i] * (log2(psum) -0.5*s*GET_Q_FOR_INDEX(i));
        // With a uniform weighting scheme, HQEnergyCorrection = -0.5*GET_Q_FOR_INDEX(i)
        for (int i=0; i<NUM_Q; ++i) workHQ[i] = HQEnergyCorrection[effectiveNLevel][i];

        DecompositionLevel *level = nextLevel;
        for (int scale = 1; scale<=effectiveNLevel; ++scale) {
            for (int i=0; i<NUM_Q; ++i) {
                double psum = (level->powerSum[i] + level->powerSumLastTerm[i]) / (level->pscount + 1.0);
#ifdef NO_YLOG2X_ASM
                workHQ[i] += powerSumToHqConstant[effectiveNLevel][scale] * log2(psum);
#else
                asm(
                "fyl2x\n\t"    // powerSumToHqConstant[effectiveNLevel][scale] * log2(psum) in one instruction
                "faddl %[hq]" // loads and add workHQ[i]
                : "=t" (workHQ[i])
                : "u" (powerSumToHqConstant[effectiveNLevel][scale]), "0" (psum), [hq] "m" (workHQ[i])
                : "st(1)"
                );
#endif
            }
            level = level->nextLevel;
        }
        // Use 0.0 so it can be added & subtracted later on, NaN cannot
        for (int i=0; i<NUM_Q; ++i) if (!isnormal(workHQ[i])) workHQ[i] = 0.0;


    }

    // process correlation
    else  {
        if (correlationHistory.capacity()==1) workSquaredCorrelation = squaredCorrelation;
        else {
            if (correlationHistory.isFull()) {
                workSquaredCorrelation = correlationHistory[0];
                for (int i=0; i<NUM_Q; ++i) squaredCorrelationSum[i] -= workSquaredCorrelation[i];
                correlationHistory.push_cyclic(workSquaredCorrelation); //pushes itself at the same place!
            } else {
                correlationHistory.push_back(workSquaredCorrelation = new double[NUM_Q]);
            }
        }

        // correction term is included below when monitoring correlations
        memset(workHQ, 0, NUM_Q * sizeof(double));
        memset(workSquaredCorrelation, 0, NUM_Q * sizeof(double));

        // Temporary place holders on the stack for each q
        double sumWY[NUM_Q] = EMPTY_Q_INITIALIZER;
        double sumWXY[NUM_Q] = EMPTY_Q_INITIALIZER;
        double sumWYY[NUM_Q] = EMPTY_Q_INITIALIZER;

        DecompositionLevel *level = nextLevel;
        for (int scale = 1; scale<=effectiveNLevel; ++scale) {
            double weight = weights[scale-1];
            for (int i=0; i<NUM_Q; ++i) {
                double psum = (level->powerSum[i] + level->powerSumLastTerm[i]) / (level->pscount + 1.0);
                double log2y;
#ifdef NO_YLOG2X_ASM
                log2y = log2(psum);
#else
                asm(
                "fld1\n"
                "fxch\n"
                "fyl2x"    // powerSumToHqConstant[scale][i] * log2(level->powerSum[i]) in one instruction
                : "=t" (log2y)
                : "0" (psum)
                : "st(1)"
                );
#endif
                log2y -= 0.5*scale*GET_Q_FOR_INDEX(i);
                workHQ[i] += powerSumToHqConstant[effectiveNLevel][scale] * log2y;
                double log2yW = log2y * weight;
                sumWY[i] += log2yW;
                sumWYY[i] += log2yW * log2y;
                sumWXY[i] += log2yW * scale;
            }
            level = level->nextLevel;
        }
        for (int i=0; i<NUM_Q; ++i) if (!isnormal(workHQ[i])) workHQ[i] = 0.0;

        // Complete the correlation coefficients computation
        for (int i=0; i<NUM_Q; ++i) {
            double cross_product = sumWXY[i] - meanX[effectiveNLevel] * sumWY[i];
            workSquaredCorrelation[i] = cross_product * cross_product / (correlationCoefficientDenominatorTerm1[effectiveNLevel] * (sumWYY[i] - sumWY[i] * sumWY[i] * sumOfWeightsInv[effectiveNLevel]));
            if (!isnormal(workSquaredCorrelation[i])) workSquaredCorrelation[i] = 0.0;
            if (correlationHistory.capacity()>1) {
                squaredCorrelationSum[i] += workSquaredCorrelation[i];
                squaredCorrelation[i] = squaredCorrelationSum[i] / correlationHistory.size();
            }
        }
    }

    // Update hq if using an history. Otherwise, hq was used directly as workHQ
    if (hqHistory.capacity()!=1) for (int i=0; i<NUM_Q; ++i) hq[i] += workHQ[i];

    // Could put a constraint on the ordering of the Q array => avoid the map
    map<double,double> qtau;
    if (hqHistory.capacity()!=1) for (int i=0; i<NUM_Q; ++i) qtau[GET_Q_FOR_INDEX(i)] = (hq[i]/(double)hqHistory.size()) - 1.0;
    else for (int i=0; i<NUM_Q; ++i) qtau[GET_Q_FOR_INDEX(i)] = hq[i] - 1.0;
    qtau[0] = -1.0;

    alpha_D.clear();

    // legendre transform

    if (secondOrderLegendreApproximation) {
        // Cleanup non-convex artifacts before legendre transform
        list< pair<double,double> > qtaulist;
        // copy in list since map iterators are awfully slow
        for(map<double,double>::const_iterator it = qtau.begin(); it!=qtau.end(); ++it) qtaulist.push_back(*it);
        bool unclean;
        do {
            unclean = false;
            list< pair<double,double> >::iterator previt = qtaulist.begin();
            if (previt==qtaulist.end()) return; // really degenerate case: empty list
            list< pair<double,double> >::iterator it = previt; ++it;
            if (it==qtaulist.end()) return; // degenerate case: one point, no spectrum
            list< pair<double,double> >::iterator nextit = it; ++nextit;
            if (nextit==qtaulist.end()) {
                // degenerate case
                double alpha = (it->second - previt->second) / (it->first - previt->first);
                double D_alpha = it->first * alpha - it->second;
                if (D_alpha>1.0) D_alpha = 1.0;
                if (D_alpha>0.0) alpha_D[alpha] = D_alpha;
                return;
            }
            // Check for non-convex qtau artifacts
            while(nextit!=qtaulist.end()) {
                double alpha = (nextit->second - previt->second) / (nextit->first - previt->first);
                double midpoint = previt->second + alpha * (it->first - previt->first);
                if (it->second < midpoint) {
                    qtaulist.erase(it);
                    unclean = true;
                } else {
                    previt = it;
                }
                it = nextit++;
            }
        } while (unclean);
        //can now use the ordered, cleaned, qtau list
        list< pair<double,double> >::iterator previt = qtaulist.begin();
        list< pair<double,double> >::iterator it = previt; ++it;
        list< pair<double,double> >::iterator nextit = it; ++nextit;
        while(nextit!=qtaulist.end()) {
            double alpha = (nextit->second - previt->second) / (nextit->first - previt->first);
            double D_alpha = it->first * alpha - it->second;
            if (D_alpha>1.0) D_alpha = 1.0;
            if (D_alpha>0.0) alpha_D[alpha] = D_alpha;
            previt = it; it = nextit++;
        }
    }
    else {
        map<double,double>::const_iterator previt = qtau.begin();
        map<double,double>::const_iterator it = previt;
        ++it;
        while(it!=qtau.end()) {
            double alpha = (it->second - previt->second) / (it->first - previt->first);
            double D_alpha = it->first * alpha - it->second;
            if (D_alpha>1.0) D_alpha = 1.0;
            if (D_alpha>0.0) alpha_D[alpha] = D_alpha;
            previt = it++;
        }
    }

}

void MultifractalAnalyzer::suggestParametersForDataSize(int dataSize, Wavelet* w, int& nlevel, int& ncoefmin, int& navg) {
    int s = w->support-2;
    if (s<2) s=2;
    nlevel = (int)(floor(log2((double)(dataSize + s)/(double)s))) - 1;
    if (nlevel<=0) {
        nlevel = 1;
        ncoefmin = s;
        navg = 1;
        return;
    }
    ncoefmin = (dataSize + s) / (1 << nlevel) - s;
    // heuristic, we could technically reconstruct the data with half the support
    // but we need enough data at the highest level to make statistics
    while (ncoefmin<w->support) {
        --nlevel;
        if (nlevel<=0) {nlevel = 1; ncoefmin = s; navg = 1; return;}
        ncoefmin = (dataSize + s) / (1 << nlevel) - s;
    }
    navg = dataSize + s - (1 << nlevel) * (ncoefmin + s);
    if (navg<1) navg = 1;
}

/*

// For monofractal estimates
// return alpha for which D_alpha = 1 or as close as possible
double getAlphaForDMax() {
    ...
}
*/
