/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (initial code from the incremfa project)              *
 *   - INRIA (modifications/diffs in SVN)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
using namespace std;

#include <assert.h>


#include "WaveletFilter.h"

WaveletFilter::WaveletFilter(int nlevel, Wavelet* w) : data(w->support), band(nlevel+1) {

    wavelet = w->ref();


    wsm2 = wavelet->support - 2;
    wsm2mwd2 = wsm2-(wavelet->support/2);
    if (wsm2mwd2<0) wsm2mwd2=0;
    initialNaN = wsm2 * (1 << nlevel) - 1;
    if (initialNaN<0) initialNaN=0;

    // special case for Haar Wavelet
    if (wsm2<2) {
        wsm2=2;
        initialNaN=1;
    }

    delay = ((1 << nlevel) - 1) * wsm2;

    nextLevel = new DecompositionLevel(nlevel-1, wsm2, w);
    nextLevelWaiting = new DecompositionLevel(nlevel-1, wsm2, w);

    this->nlevel = nlevel;

    newDataDWTIdx = data.capacity() - wavelet->support;

    // Start with the initialization version
    // Then switch to full-buffer regime when this happens
    newDataDWTPtr = &WaveletFilter::newDataDWTInitPhase;

    for (int i=0; i<=nlevel; ++i) band[i] = 1;

}

WaveletFilter::~WaveletFilter() {

    // cleanup levels (recursive)
    delete nextLevel;
    delete nextLevelWaiting;

    // release shared instance
    wavelet->unref();
}

void WaveletFilter::reset() {

    // Clear data
    data.size() = 0;

    // Start with the initialization version
    // Then switch to full-buffer regime when this happens
    newDataDWTPtr = &WaveletFilter::newDataDWTInitPhase;

    // recursive reset levels
    if (nextLevel) nextLevel->reset();
    if (nextLevelWaiting) nextLevelWaiting->reset();
}


DecompositionLevel* WaveletFilter::newDataDWTInitPhase(double newData) {
    if (data.isFull()) {
        // remove old data
        data.pop_front();
        // Switch to full buffer version for next time
        newDataDWTPtr = &WaveletFilter::newDataDWTFullRegime;
    }
    // add new data
    data.push_back(newData);
    // Swap buffers
    DecompositionLevel* tmplp = nextLevelWaiting;
    nextLevelWaiting = nextLevel;
    nextLevel = tmplp;
    // update DWT coefficients for feeding next level recursively
    const int idx = data.size() - wavelet->support;
    if (idx<0) return 0;
    double lphp[2];
    wavelet->convolve(data, idx, lphp);
    return nextLevel->newDataUpdate(lphp);
}

DecompositionLevel* WaveletFilter::newDataDWTFullRegime(double newData) {
    // add new data
    data.push_cyclic(newData);
    // Swap buffers
    DecompositionLevel* tmplp = nextLevelWaiting;
    nextLevelWaiting = nextLevel;
    nextLevel = tmplp;
    // update DWT coefficients for feeding next level recursively
    double lphp[2];
    wavelet->convolve(data, newDataDWTIdx, lphp);
    return nextLevel->newDataUpdate(lphp);
}


int WaveletFilter::filter(double data, double& filtered) {

    // take care of initialization phase.
    DecompositionLevel *lastLevel = newDataDWT(data);
    if (!lastLevel) return 0;
    if (!lastLevel->lowPass.isFull()) return 0;

    //assert(lastLevel->lowPass.size() == wavelet->support - 2);

    // Initialize with either low-pass coefs or null at high-level
    CyclicArray* topSignal = &lastLevel->lowPass;
    CyclicArray dummy(wsm2);
    if (!band[nlevel]) {
        dummy.zeroFill();
        topSignal = &dummy;
    }

    CyclicArray filteredData(wsm2);

    // top level a bit special as we must handle the two bands
    if (band[nlevel] && band[nlevel-1]) for (int i=0; i<=wsm2mwd2; ++i) {
        wavelet->reconstruct(lastLevel->lowPass, lastLevel->highPass, i, filteredData.array+i*2);
    }
    else if (band[nlevel]) for (int i=0; i<=wsm2mwd2; ++i) {
        wavelet->smooth(lastLevel->lowPass, i, filteredData.array+i*2);
    }
    else if (band[nlevel-1]) for (int i=0; i<=wsm2mwd2; ++i) {
        wavelet->delta(lastLevel->highPass, i, filteredData.array+i*2);
    }
    else filteredData.zeroFill();

    int bandN = nlevel-2;
    int hpshift = wsm2;
    // filteredData reconstructed with band filter. Now get down to data level
    for (DecompositionLevel *level = lastLevel->prevLevel; level; level = level->prevLevel) {
        CyclicArray reconstructedData(wsm2);
        int startHP = level->highPass.size() - hpshift - wsm2;
        if (band[bandN]) for (int i=0; i<=wsm2mwd2; ++i) {
            wavelet->reconstruct(filteredData, level->highPass, i, startHP+i, reconstructedData.array+i*2);
        }
        else for (int i=0; i<=wsm2mwd2; ++i) {
            wavelet->smooth(filteredData, i, reconstructedData.array+i*2);
        }
        filteredData.swap(reconstructedData);
        //filteredData.size() = wsm2;
        --bandN;
        hpshift = hpshift * 2 + wsm2;
    }

    // take only the most recent one for the filter.
    // next frame will reconstruct next data
    filtered = filteredData[wsm2-1];

    return 1;
}

