/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (initial code from the incremfa project)              *
 *   - INRIA (modifications/diffs in SVN)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef WF_ANALYZER_H
#define WF_ANALYZER_H

#include <vector>

#include "DecompositionLevel.h"

/// Final structure is a vector of data + its decomposition
/// Name based on Dynamic Multi-Fractal Analysis with Discrete Wavelet Transform
/// Suggested names like DMFADWT, diwtradymfrana... Too complex
/// Let's name it something simple
class WaveletFilter {
protected:

    // pointer to member type
    typedef DecompositionLevel* (WaveletFilter::* NewDataDWTFunctionPtr)(double);
    NewDataDWTFunctionPtr newDataDWTPtr;

    /// Updates the DWT decomposition, takes care of non-full buffer effects
    /// and automatically switched to the full buffer version when this happens
    DecompositionLevel* newDataDWTInitPhase(double newData);

    /// Faster version of the DWT update, assuming the buffer is full
    DecompositionLevel* newDataDWTFullRegime(double newData);
    int newDataDWTIdx;

    int wsm2mwd2;
    int wsm2;
public:

    int nlevel;

    /// In any method using a spectrum there is a delay to get low-frequency components.
    /// This is completely independent from this algorithm, but a general fact.
    /// In our case, the delay corresponds to the difference between the newly arrived point
    /// and the last point that can be reconstructed using the DWT.
    /// In practice, it is (2^L - 1)(wavelet->support - 2)
    int delay;
    int initialNaN;



    DecompositionLevel *nextLevel;
    DecompositionLevel *nextLevelWaiting;

    /// Original data: the low pass for next level
    CyclicArray data;

    /// Analyzing wavelet
    Wavelet* wavelet;
    
    /** Public constructor takes these arguments:
        @param nlevel    Number of levels of decomposition
        @param w         The wavelet
        @note  The data size is 2^nlevel * (w->support-1) + (w->support - 2) * (2^nlevel - 1), with ^ = exponentiation
    */
    WaveletFilter(int nlevel, Wavelet* w);

    ~WaveletFilter();

    /// Update DWT for this new data.
    inline DecompositionLevel* newDataDWT(double newData) {
        return (this->*newDataDWTPtr)(newData);
    }

    /// Calls update DWT internally, then apply the filter
    /// and reconstruct the data after the delay
    /// @param data The new data value.
    /// @param filtered The filtered value after the delay lag necessary to take into account the low frequencies
    /// @return 0 If there are not yet enough data values. 1 if the delay has been reached and filtered values may be computed
    int filter(double data, double& filtered);

    /// Reset this analyzer. Use this instead of destroying/reallocating this object
    /// with the same parameters: the result is the same, but this function is much
    /// faster
    void reset();
    
    std::vector<int> band;
};

#endif
