/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu (some code taken from the drrn2 project)              *
 *   - INRIA (modifications/diffs in SVN)                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <cstdlib>
using namespace std;

#include "WaveletFilter.h"

const int default_nlevel = 6;
const string default_waveletName = "DW4";

void help(const char* progname) {
cout<<
"Usage: " << progname << " [options] input_filename output_filename\n"
"Options:\n"
"-L <int>: Number of decomposition levels (default=" << default_nlevel << ").\n"
"-B <bool,bool,...>: The (L+1) bands to include, bool = 0 or 1 (default=1,1,...1). The first L bands corresponds to including the details of the first L decompositions, the L+A band corresponds to including the remaining scaled down average values.\n"
"-W <string>: The wavelet to use (default="<<default_waveletName<<"). Available wavelets: DW4, DW6. For Daubechies Wavelet 4 and 6 respectively.\n"
"-n: Don't fill uncomputable values with NaN.\n"
"-q: Quiet mode.\n"
"\n"
"Example: " << progname << " -L3 -B0,1,1,0\n"
"This will decompose the data using 3 levels of wavelet decomposition. Then, the average part is left out (low frequencies filtered out), some details are included (medium frequencies included), but the lowest details are not (high frequencies filtered out) \n"
"Example2: " << progname << " -L5 -B0,1,1,1,1,1\n"
"This will decompose the data using 5 levels of wavelet decomposition. Then, high frequencies are filtered out.\n"
<<flush;
}

int main(int argc, char** argv) {

    bool fillNaN = true;
    bool quiet = false;
    int nlevel = default_nlevel;
    string waveletName = default_waveletName;
    vector<int> bands;

    int c; opterr = 0;
    while ((c=getopt(argc,argv,"L:B:W:nq"))!=-1) switch(c) {
        case 'L': nlevel = atoi(optarg); break;
        case 'B': {
            char *c = optarg;
            while (*c) {if (*c==',') *c= ' '; ++c;}
            stringstream tokenizer(optarg);
            int band;
            while (tokenizer >> band) bands.push_back(band);
            break;
        }
        case 'W': waveletName = optarg; break;
        case 'n': fillNaN = false; break;
        case 'q': quiet = true; break;
        default: help(argv[0]); return 0;
    }

    if (nlevel<=0) {
        cout << "Invalid number of levels" << endl;
        help(argv[0]);
        return 0;
    }

    if (bands.empty()) {
        bands.reserve(nlevel+1);
        for (int i=0; i<=nlevel; ++i) bands.push_back(1);
    }

    if (bands.size()!=(unsigned int)nlevel+1) {
        cout << "Invalid number of bands" << endl;
        help(argv[0]);
        return 0;
    }
    bool bandsOK = true;
    for (int i=0; i<=nlevel; ++i) if (bands[i]!=0 && bands[i]!=1) {bandsOK = false; break;}
    if (!bandsOK) {
        cout << "Invalid band filter:";
        for (int i=0; i<=nlevel; ++i) cout << " " << bands[i];
        cout << endl;
        help(argv[0]);
        return 0;
    }

    Wavelet *wavelet = 0;
    if (waveletName=="DW4") wavelet = DW4::getInstance();
    else if (waveletName=="DW6") wavelet = DW6::getInstance();
    else if (waveletName=="HW") wavelet = HW::getInstance();
    else {
        cout << "Invalid wavelet name: " << waveletName << endl;
        help(argv[0]);
        return 0;
    }

    if (argc - optind < 2) {
        help(argv[0]);
        return 0;
    }

    ifstream inputfile(argv[optind]);
    if (!inputfile) {
        cout << "Cannot read input file: " << argv[optind] << endl;
        return 0;
    }

    ofstream outputfile(argv[optind+1]);
    if (!outputfile) {
        cout << "Cannot write to output file: " << argv[optind+1] << endl;
        return 0;
    }

    if (!quiet) {
        cout << "Processing file '" << argv[optind] << "' into '" << argv[optind+1] << "'" << endl;
        cout << "Using wavelet " << waveletName << endl;
        cout << "With "<<nlevel<<" levels of decomposition" << endl;
        cout << "Band filter:";
        for (int i=0; i<=nlevel; ++i) cout << " " << bands[i];
        cout << endl;
    }

    vector<WaveletFilter*> filters;
    string line; double value, filtered;


    while (inputfile && !inputfile.eof()) {
        getline(inputfile, line);
        if (line.empty()) continue;
        stringstream linereader(line);
        if (filters.empty()) {
            while (linereader >> value) {
                filters.push_back(new WaveletFilter(nlevel,wavelet));
                for (int i=0; i<=nlevel; ++i) filters.back()->band[i] = bands[i];
                filters.back()->filter(value, filtered);
            }
            if (!quiet) {
                cout << filters[0]->initialNaN << " values cannot be reconstructed at the beginning of the series" << endl;
                cout << filters[0]->delay << " values cannot be reconstructed at the end of the series" << endl;
            }
            if (fillNaN) for (int i=0; i<filters[0]->initialNaN; ++i) {
                for (unsigned int j=0; j<filters.size(); ++j) {
                    if (j) outputfile << " ";
                    outputfile << "NaN";
                }
                outputfile << endl;
            }
        } else {
            vector<WaveletFilter*>::iterator it = filters.begin();
            vector<WaveletFilter*>::iterator eit = filters.end();
            int res = 0;
            while (linereader >> value) {
                if (it==eit) break;
                res = (*it)->filter(value, filtered);
                if (res) {
                    if (it!=filters.begin()) outputfile << " ";
                    outputfile << filtered;
                }
                ++it;
            }
            if (res) outputfile << endl;
        }
    }

    if (fillNaN) for (int i=0; i<filters[0]->delay; ++i) {
        for (unsigned int j=0; j<filters.size(); ++j) {
            if (j) outputfile << " ";
            outputfile << "NaN";
        }
        outputfile << endl;
    }

    inputfile.close();
    outputfile.close();
}
